import logging

from django.db import connections, models

logger = logging.getLogger(__name__)


class ActivityManager(models.Manager):
    """Model management for worker activities"""

    def bulk_insert(self, worker_version_id, process_id, elements_qs, state=None):
        """
        Create initial worker activities from a queryset of elements in a efficient way.
        Due to the possible large amount of elements, we use a bulk insert from the elements query (best performances).
        The `ON CONFLICT` clause allows to automatically skip elements that already have an activity with this version.
        """
        from arkindex.dataimport.models import WorkerActivityState

        if state is None:
            state = WorkerActivityState.Queued

        assert isinstance(state, WorkerActivityState), 'State should be an instance of WorkerActivityState'

        sql, params = elements_qs.values('id').query.sql_with_params()

        with connections['default'].cursor() as cursor:
            cursor.execute(
                f"""
                INSERT INTO dataimport_workeractivity
                    (element_id, worker_version_id, state, process_id, id, created, updated)
                SELECT
                    elt.id,
                    '{worker_version_id}'::uuid,
                    '{state.value}',
                    '{process_id}',
                    uuid_generate_v4(),
                    current_timestamp,
                    current_timestamp
                FROM ({sql}) AS elt
                ON CONFLICT (element_id, worker_version_id) DO NOTHING
                """,
                params
            )


class CorpusWorkerVersionManager(models.Manager):

    def rebuild(self):
        """
        Rebuild the corpus worker versions cache from all ML results.
        """
        from arkindex.documents.models import Element, Transcription, Entity, TranscriptionEntity, Classification, MetaData

        querysets = [
            Element.objects.exclude(worker_version_id=None).values_list('corpus_id', 'worker_version_id'),
            Transcription.objects.exclude(worker_version_id=None).values_list('element__corpus_id', 'worker_version_id'),
            Entity.objects.exclude(worker_version_id=None).values_list('corpus_id', 'worker_version_id'),
            TranscriptionEntity.objects.exclude(worker_version_id=None).values_list('entity__corpus_id', 'worker_version_id'),
            Classification.objects.exclude(worker_version_id=None).values_list('element__corpus_id', 'worker_version_id'),
            MetaData.objects.exclude(worker_version_id=None).values_list('element__corpus_id', 'worker_version_id'),
        ]

        for i, queryset in enumerate(querysets, start=1):
            logger.info(f'Rebuilding cache from {queryset.model.__name__} ({i}/{len(querysets)})')
            self.bulk_create([
                self.model(corpus_id=corpus_id, worker_version_id=worker_version_id)
                for corpus_id, worker_version_id in queryset.distinct()
            ], ignore_conflicts=True)
