from typing import Optional

from django.conf import settings
from django.db import transaction
from django_rq import job

from arkindex.dataimport.models import ActivityState, DataImport, WorkerActivity, WorkerActivityState


@job('default', timeout=settings.RQ_TIMEOUTS['initialize_activity'])
def initialize_activity(process: DataImport):
    """
    List all worker versions used in a process and initialize their activity on processed elements.
    Timeout is set to 1 hour
    """
    try:
        with transaction.atomic():
            for version_id in process.versions.values_list('id', flat=True):
                WorkerActivity.objects.bulk_insert(
                    worker_version_id=version_id,
                    process_id=process.id,
                    elements_qs=process.list_elements()
                )
    except Exception as e:
        process.activity_state = ActivityState.Error
        process.save()
        raise e
    process.activity_state = ActivityState.Ready
    process.save()


@job('default', timeout=settings.RQ_TIMEOUTS['process_delete'])
def process_delete(process: DataImport, delete_activity: Optional[WorkerActivityState] = None):
    """
    Delete a process and removes its reference from related worker activities.
    If `delete` is defined, activities of this state will be deleted.
    """
    if delete_activity:
        process.activities.filter(state=delete_activity).delete()
    process.delete()
