import uuid
from unittest.mock import MagicMock, call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import (
    ActivityState,
    DataImport,
    DataImportMode,
    WorkerActivity,
    WorkerActivityState,
    WorkerVersion,
)
from arkindex.dataimport.tasks import initialize_activity
from arkindex.documents.models import Classification, ClassificationState, Element, MLClass
from arkindex.project.tests import FixtureTestCase
from arkindex.users.models import User


class TestWorkerActivity(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')
        cls.element = Element.objects.get(name='Volume 1, page 2r')
        cls.process = DataImport.objects.create(
            mode=DataImportMode.Workers,
            creator=cls.user
        )

    def setUp(self):
        # Create a queued activity for this element
        self.activity = self.element.activities.create(
            process=self.process,
            worker_version=self.worker_version,
            state=WorkerActivityState.Queued
        )

    def test_bulk_insert_activity_children(self):
        """
        Bulk insert worker activities for acts
        """
        elements_qs = Element.objects.filter(type__slug='act', type__corpus_id=self.corpus.id)
        params = {
            'worker_version_id': self.worker_version.id,
            'corpus_id': self.corpus.id,
            'state': WorkerActivityState.Started.value,
            'process_id': self.process.id,
        }
        with self.assertExactQueries('workeractivity_bulk_insert.sql', params=params):
            WorkerActivity.objects.bulk_insert(self.worker_version.id, self.process.id, elements_qs, state=WorkerActivityState.Started)
        self.assertEqual(elements_qs.count(), 5)
        self.assertEqual(WorkerActivity.objects.filter(state=WorkerActivityState.Started, process=self.process).count(), 5)

    def test_bulk_insert_activity_existing(self):
        """
        Elements in the queryset should be skipped if they already have an activity
        The old process ID is preserved during the bulk create
        """
        elements_qs = Element.objects.filter(type__slug='act', type__corpus_id=self.corpus.id)
        old_process = DataImport.objects.create(mode=DataImportMode.Workers, creator=self.user)
        WorkerActivity.objects.bulk_create([
            WorkerActivity(
                element=element,
                worker_version=self.worker_version,
                state=WorkerActivityState.Processed.value,
                process=old_process
            )
            for element in elements_qs[:2]
        ])
        with self.assertNumQueries(1):
            WorkerActivity.objects.bulk_insert(self.worker_version.id, self.process.id, elements_qs, state=WorkerActivityState.Started)
        self.assertEqual(WorkerActivity.objects.filter(element_id__in=elements_qs.values('id')).count(), 5)
        self.assertEqual(elements_qs.count(), 5)
        # Only 3 acts have been marked as started for this worker
        self.assertEqual(WorkerActivity.objects.filter(state=WorkerActivityState.Started).count(), 3)
        self.assertEqual(WorkerActivity.objects.filter(process=old_process).count(), 2)

    @patch('arkindex.project.triggers.dataimport_tasks.initialize_activity.delay')
    def test_bulk_insert_children_class_filter(self, activities_delay_mock):
        """
        Worker activities creation should work with complex elements selection (e.g. with a class filter)
        """
        # Mock workers activity to make it a synchronous job
        from arkindex.dataimport.tasks import initialize_activity
        activities_delay_mock.side_effect = initialize_activity

        agent_class = MLClass.objects.create(name='James', corpus=self.corpus)
        Classification.objects.bulk_create(
            Classification(ml_class=agent_class, state=ClassificationState.Validated, element=e)
            for e in self.corpus.elements.filter(type__slug='page')
        )
        dataimport = self.corpus.imports.create(
            activity_state=ActivityState.Pending,
            creator=self.user,
            mode=DataImportMode.Workers,
            corpus=self.corpus,
            best_class=agent_class.name
        )
        dataimport.worker_runs.create(version=self.worker_version, parents=[])
        with self.assertNumQueries(22):
            dataimport.start()

        self.assertCountEqual(
            WorkerActivity.objects.filter(worker_version=self.worker_version).values_list('element_id', flat=True),
            self.corpus.elements.filter(type__slug='page').values_list('id', flat=True)
        )

    def test_put_activity_requires_internal(self):
        """
        Only internal users (workers) are able to update the state of a worker activity
        Internal users with an instance admin are able to update a worker activity
        """
        internal_admin_user = User.objects.create_user('god@test.test', 'G0D')
        internal_admin_user.is_internal = True
        internal_admin_user.is_admin = True
        internal_admin_user.save()
        cases = (
            (None, status.HTTP_403_FORBIDDEN, 0),
            (self.user, status.HTTP_403_FORBIDDEN, 2),
            (self.superuser, status.HTTP_403_FORBIDDEN, 2),
            (self.internal_user, status.HTTP_200_OK, 4),
            (internal_admin_user, status.HTTP_200_OK, 4)
        )
        for user, status_code, requests_count in cases:
            self.activity.state = WorkerActivityState.Queued
            self.activity.save()
            if user:
                self.client.force_login(user)
            with self.assertNumQueries(requests_count):
                response = self.client.put(
                    reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                    {
                        'element_id': str(self.element.id),
                        'process_id': str(self.process.id),
                        'state': WorkerActivityState.Started.value,
                    },
                    content_type='application/json',
                )
            self.assertEqual(response.status_code, status_code)

    def test_put_activity_wrong_worker_version(self):
        """
        Raises a generic error in case the worker version does not exists because a single SQL request is performed
        The response is a HTTP_409_CONFLICT
        """
        self.client.force_login(self.internal_user)
        with self.assertNumQueries(4):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(uuid.uuid4())}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(self.process.id),
                    'state': WorkerActivityState.Started.value,
                },
                content_type='application/json',
            )
        self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)
        self.assertDictEqual(response.json(), {
            '__all__': [
                'Either this worker activity does not exists or '
                f"updating the state to '{WorkerActivityState.Started.value}' is forbidden."
            ]
        })

    def test_put_activity_element_unexisting(self):
        """
        Raises a generic error in case no activity exists for this element
        The response is a HTTP_409_CONFLICT
        """
        self.client.force_login(self.internal_user)
        with self.assertNumQueries(4):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(uuid.uuid4()),
                    'process_id': str(self.process.id),
                    'state': WorkerActivityState.Started.value,
                },
                content_type='application/json',
            )
        self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)
        self.assertDictEqual(response.json(), {
            '__all__': [
                'Either this worker activity does not exists or '
                f"updating the state to '{WorkerActivityState.Started.value}' is forbidden."
            ]
        })

    def test_put_activity_process_unexisting(self):
        """
        Raises an error in case the process does not exist
        """
        self.client.force_login(self.internal_user)
        wrong_process_id = uuid.uuid4()
        with self.assertNumQueries(3):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(wrong_process_id),
                    'state': WorkerActivityState.Started.value,
                },
                content_type='application/json',
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'process_id': [f'Invalid pk "{wrong_process_id}" - object does not exist.']
        })

    def test_put_activity_allowed_states(self):
        """
        Check each case state update is allowed depending on the actual state of the activity
        """
        queued, started, error, processed = (
            WorkerActivityState[state].value
            for state in ['Queued', 'Started', 'Error', 'Processed']
        )
        allowed_states_update = (
            (queued, started),
            (started, processed),
            (started, error),
            (error, queued),
            (error, started),
        )
        self.client.force_login(self.internal_user)
        for state, payload_state in allowed_states_update:
            self.activity.state = state
            self.activity.save()
            with self.assertNumQueries(4):
                response = self.client.put(
                    reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                    {
                        'element_id': str(self.element.id),
                        'process_id': str(self.process.id),
                        'state': payload_state,
                    },
                    content_type='application/json',
                )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.activity.refresh_from_db()
            self.assertEqual(self.activity.state.value, payload_state)

    def test_put_activity_forbidden_states(self):
        """
        Check state update is forbidden for some non consistant cases
        The response is a HTTP_409_CONFLICT
        """
        queued, started, error, processed = (
            WorkerActivityState[state].value
            for state in ['Queued', 'Started', 'Error', 'Processed']
        )
        forbidden_states_update = (
            (queued, queued),
            (started, queued),
            (queued, error),
            (processed, error),
        )
        self.client.force_login(self.internal_user)
        for state, payload_state in forbidden_states_update:
            self.activity.state = state
            self.activity.save()
            with self.assertNumQueries(4):
                response = self.client.put(
                    reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                    {
                        'element_id': str(self.element.id),
                        'process_id': str(self.process.id),
                        'state': payload_state,
                    },
                    content_type='application/json',
                )
            self.assertEqual(response.status_code, status.HTTP_409_CONFLICT)
            self.assertDictEqual(response.json(), {
                '__all__': [
                    'Either this worker activity does not exists or '
                    f"updating the state to '{payload_state}' is forbidden."
                ]
            })
            self.activity.refresh_from_db()
            self.assertEqual(self.activity.state.value, state)

    def test_put_activity_process_overriding(self):
        """
        A process can update a worker activity event if it was initialized in another process
        """
        self.client.force_login(self.internal_user)
        process2 = DataImport.objects.create(
            mode=DataImportMode.Workers,
            creator=self.user
        )
        self.activity.process_id = process2.id
        self.activity.save()

        with self.assertNumQueries(4):
            response = self.client.put(
                reverse('api:update-worker-activity', kwargs={'pk': str(self.worker_version.id)}),
                {
                    'element_id': str(self.element.id),
                    'process_id': str(process2.id),
                    'state': WorkerActivityState.Started.value,
                },
                content_type='application/json',
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'element_id': str(self.element.id),
            'process_id': str(process2.id),
            'state': WorkerActivityState.Started.value,
        })
        self.activity.refresh_from_db()
        self.assertEqual(self.activity.process_id, process2.id)

    @patch('arkindex.dataimport.models.DataImport.versions')
    @patch('arkindex.dataimport.models.ActivityManager.bulk_insert')
    def test_async_activities_error(self, bulk_insert_mock, versions_mock):
        """
        Handles errors while initializing workers activity of a process
        """
        process = self.corpus.imports.create(
            creator=self.user,
            mode=DataImportMode.Workers,
            element_type=self.corpus.types.get(slug='volume'),
            activity_state=ActivityState.Pending
        )
        # Mock list_elements method to return the same queryset object
        elts_qs = Element.objects.filter(type__slug='volume')
        self.assertEqual(elts_qs.count(), 2)
        process.list_elements = MagicMock()
        process.list_elements.return_value = elts_qs

        v1_id, v2_id = uuid.uuid4(), uuid.uuid4()
        bulk_insert_mock.side_effect = [
            bulk_insert_mock,
            Exception('Something went wrong')
        ]

        process.versions = MagicMock()
        process.versions.values_list.return_value = [v1_id, v2_id]

        with self.assertRaises(Exception):
            initialize_activity(process)

        self.assertEqual(bulk_insert_mock.call_count, 2)
        self.assertListEqual(bulk_insert_mock.call_args_list, [
            call(worker_version_id=v1_id, process_id=process.id, elements_qs=elts_qs),
            call(worker_version_id=v2_id, process_id=process.id, elements_qs=elts_qs)
        ])

        process.refresh_from_db()
        self.assertEqual(process.activity_state, ActivityState.Error)
        self.assertEqual(WorkerActivity.objects.filter(element__type__slug='volume').count(), 0)
