import itertools
import uuid

from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import DataImport, DataImportMode, WorkerActivity, WorkerActivityState, WorkerVersion
from arkindex.documents.models import Corpus
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User


class TestWorkersActivity(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.version_1 = WorkerVersion.objects.get(worker__slug='reco')
        cls.version_2 = WorkerVersion.objects.get(worker__slug='dla')
        cls.private_corpus = Corpus.objects.create(name='private', public=False)
        cls.process = DataImport.objects.create(
            mode=DataImportMode.Workers,
            creator=cls.user,
            corpus=cls.corpus,
        )

        # Generate worker activities
        WorkerActivity.objects.bulk_create([
            *(
                WorkerActivity(
                    element_id=elt.id,
                    state=state,
                    worker_version_id=cls.version_1.id,
                    process_id=cls.process.id,
                ) for elt, state in zip(cls.corpus.elements.all(), itertools.cycle(WorkerActivityState))
            ), *(
                WorkerActivity(
                    element_id=elt.id,
                    state=WorkerActivityState.Processed.value,
                    worker_version_id=cls.version_2.id,
                    process_id=None,
                ) for elt in cls.corpus.elements.all()
            )
        ])
        cls.error, cls.processed, cls.queued, cls.started = [
            WorkerActivity.objects.filter(
                element__corpus_id=cls.corpus.id,
                worker_version_id=cls.version_1.id,
                state=state
            ).count()
            for state in [
                WorkerActivityState.Error,
                WorkerActivityState.Processed,
                WorkerActivityState.Queued,
                WorkerActivityState.Started
            ]
        ]

    def test_workeractivities_stats_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse('api:corpus-workers-activity', kwargs={'corpus': str(self.corpus.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_workeractivities_private_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:corpus-workers-activity', kwargs={'corpus': str(self.private_corpus.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have guest access to this corpus.'})

    def test_workeractivities_unexisting_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse('api:corpus-workers-activity', kwargs={'corpus': str(uuid.uuid4())})
            )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_workeractivities_empty(self):
        """
        Handle a corpus that has absolutely no activity
        A user with a guest access can retrieve statistics
        """
        user = User.objects.create_user('user42@test.test', 'abcd')
        user.verified_email = True
        user.save()
        self.private_corpus.memberships.create(user=user, level=Role.Guest.value)
        self.client.force_login(user)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:corpus-workers-activity', kwargs={'corpus': str(self.private_corpus.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json(), [])

    def test_workers_activity_distributed_states(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-workers-activity', kwargs={'corpus': str(self.corpus.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(response.json(), [
            {
                'worker_version_id': str(self.version_1.id),
                'queued': self.queued,
                'started': self.started,
                'processed': self.processed,
                'error': self.error,
            }, {
                'worker_version_id': str(self.version_2.id),
                'queued': 0,
                'started': 0,
                'processed': self.corpus.elements.count(),
                'error': 0,
            }
        ])

    def test_process_activity_stats_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse('api:process-workers-activity', kwargs={'pk': str(self.process.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_process_activity_stats_private(self):
        self.process.corpus = self.private_corpus
        self.process.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(
                reverse('api:process-workers-activity', kwargs={'pk': str(self.process.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have an admin access to this process.'})

    def test_process_activity_stats_unexisting(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse('api:process-workers-activity', kwargs={'pk': str(uuid.uuid4())})
            )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_process_activity_stats(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(
                reverse('api:process-workers-activity', kwargs={'pk': str(self.process.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json(), [
            {
                'worker_version_id': str(self.version_1.id),
                'queued': self.queued,
                'started': self.started,
                'processed': self.processed,
                'error': self.error,
            }
        ])
