#!/usr/bin/env python3
import uuid

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from arkindex.documents.indexer import Indexer
from arkindex.documents.models import Corpus


class Command(BaseCommand):
    help = 'Reindex elements, transcriptions and classifications into Solr'

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "corpus_id", help="UUID of an existing corpus to reindex", type=uuid.UUID
        )
        parser.add_argument(
            '--drop',
            help="Drop the existing indexes before reindexing",
            action='store_true',
        )
        parser.add_argument(
            '--setup',
            help="Only setup a collection. Create a collection and fields if they do not exist or update the fields",
            action='store_true',
        )

    def handle(self, corpus_id, **options):
        if not settings.ARKINDEX_FEATURES['search']:
            raise CommandError('Reindexation is not possible if the search feature flag is disabled. '
                               'Consider setting `features.search` to `on` or `true` or `yes` in the YAML '
                               'configuration file, and configuring Solr properly.')

        try:
            corpus = Corpus.objects.get(id=corpus_id)
        except Corpus.DoesNotExist:
            raise CommandError(f'Corpus {corpus_id} does not exist')

        if not corpus.indexable:
            raise CommandError(f'Corpus {corpus.name} is not indexable')

        indexer = Indexer(corpus.id)
        indexer.setup()
        if options.get('setup'):
            return
        if options.get('drop'):
            indexer.drop_index()
        indexer.index()
