import io
import re
from unittest.mock import patch

from django.core.management import call_command

from arkindex.project.tests import FixtureTestCase


class TestTelegraf(FixtureTestCase):
    """Tests for telegraf reporting CLI"""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.influx_re = re.compile(r'^(\w+)(,[\w\-_,=]+)? ([\w\-_,=]+) (\d+)$')
        cls.influx_measurements_re = re.compile(r'(\w+)=(\d+)')
        cls.influx_tags_re = re.compile(r'(\w+)=([\w\-]+)')
        cls.expected_output = {
            'images': {
                'total': 6,
                'checked': 0,
                'errors': 0,
                'tags': {'imageserver': 'Test_Server_1'},
            }
        }

    def _parse_influx_lines(self, lines):
        data = {}
        for line in lines:
            details = self.influx_re.search(line)
            self.assertIsNotNone(details)
            name, tags, values, t = details.groups()
            self.assertTrue(int(t) > 0)
            data[name] = {
                k: int(v)
                for k, v in self.influx_measurements_re.findall(values)
            }
            data[name]["tags"] = {
                k: v
                for k, v in self.influx_tags_re.findall(tags)
            } if tags else {}
        return data

    def test_command(self):

        output = io.StringIO()
        call_command('telegraf', stdout=output)
        self.assertIsNotNone(output)

        lines = list(filter(None, output.getvalue().split('\n')))
        self.assertEqual(len(lines), 1)

        self.assertDictEqual(
            self._parse_influx_lines(lines),
            self.expected_output,
        )

    @patch('arkindex.documents.management.commands.telegraf.requests')
    def test_post_influxdb(self, requests_mock):
        output = io.StringIO()
        call_command(
            'telegraf',
            influxdb='http://somewhere:8086/write',
            stdout=output,
        )
        self.assertIsNotNone(output)
        lines = list(filter(None, output.getvalue().split('\n')))

        self.assertDictEqual(
            self._parse_influx_lines(lines),
            self.expected_output,
        )

        self.assertEqual(requests_mock.post.call_count, 1)
        args, kwargs = zip(*requests_mock.post.call_args_list)

        # All requests are to the same URL
        self.assertSetEqual(set(args), {('http://somewhere:8086/write', )})

        self.assertSequenceEqual(kwargs, [
            {'params': {'db': 'arkindex'}, 'data': line}
            for line in lines
        ])

        # Raise exceptions on each POST request
        self.assertEqual(requests_mock.post().raise_for_status.call_count, 1)
