import datetime
from unittest.mock import call, patch

from django.contrib.auth.models import AnonymousUser
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element
from arkindex.images.models import Image
from arkindex.project.default_corpus import DEFAULT_CORPUS_TYPES
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User

FAKE_NOW = datetime.datetime.now()
# Fake DB fixtures creation date
DB_CREATED = '2020-02-02T01:23:45.678000Z'

EXPECTED_CORPUS_TYPES = [
    {
        'slug': 'volume',
        'display_name': 'Volume',
        'folder': True,
    },
    {
        'slug': 'page',
        'display_name': 'Page',
        'folder': False,
    },
    {
        'slug': 'act',
        'display_name': 'Act',
        'folder': False,
    },
    {
        'slug': 'surface',
        'display_name': 'Surface',
        'folder': False,
    },
    {
        'slug': 'text_line',
        'display_name': 'Line',
        'folder': False,
    },
    {
        'slug': 'word',
        'display_name': 'Word',
        'folder': False,
    }
]


class TestCorpus(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.anon = AnonymousUser()

        cls.corpus_public = Corpus.objects.get(name='Unit Tests', public=True)
        with patch('django.utils.timezone.now') as mock_now:
            mock_now.return_value = FAKE_NOW
            cls.corpus_private = Corpus.objects.create(name='B Private')
        cls.corpus_private.memberships.create(user=cls.user, level=Role.Contributor.value)
        cls.corpus_admin = User.objects.create(email='another_user@user.fr', verified_email=True)
        cls.corpus_private.memberships.create(user=cls.corpus_admin, level=Role.Admin.value)
        with patch('django.utils.timezone.now') as mock_now:
            mock_now.return_value = FAKE_NOW
            cls.corpus_hidden = Corpus.objects.create(name='C Hidden')

    def test_anon(self):
        # An anonymous user has only access to public
        response = self.client.get(reverse('api:corpus'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()

        self.assertEqual(len(data), 1)
        # Ignore ordering on types and type id
        types = data[0].pop('types')
        for t in types:
            del t['id']
        self.assertCountEqual(types, EXPECTED_CORPUS_TYPES)

        self.assertListEqual(
            data,
            [
                {
                    'id': str(self.corpus_public.id),
                    'rights': ['read'],
                    'public': True,
                    'indexable': False,
                    'name': 'Unit Tests',
                    'description': '',
                    'created': DB_CREATED,
                    'authorized_users': 1,
                    'thumbnail': None,
                    'thumbnail_url': None,
                    'top_level_type': None,
                }
            ]
        )

    def test_user(self):
        # A normal user has access to public + its private
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:corpus'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()

        self.assertEqual(len(data), 2)
        self.assertListEqual(data[0].pop('types'), [])
        types = data[1].pop('types')
        for t in types:
            del t['id']
        self.assertCountEqual(types, EXPECTED_CORPUS_TYPES)

        self.assertListEqual(
            data,
            [
                {
                    'id': str(self.corpus_private.id),
                    'rights': ['read', 'write'],
                    'public': False,
                    'indexable': False,
                    'name': 'B Private',
                    'description': '',
                    'created': str(FAKE_NOW).replace(' ', 'T') + 'Z',
                    'authorized_users': 2,
                    'thumbnail': None,
                    'thumbnail_url': None,
                    'top_level_type': None,
                },
                {
                    'id': str(self.corpus_public.id),
                    'rights': ['read', 'write', 'admin'],
                    'public': True,
                    'indexable': False,
                    'name': 'Unit Tests',
                    'description': '',
                    'created': DB_CREATED,
                    'authorized_users': 1,
                    'thumbnail': None,
                    'thumbnail_url': None,
                    'top_level_type': None,
                }
            ]
        )

    def test_admin(self):
        # An admin has access to everything
        self.client.force_login(self.superuser)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:corpus'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()

        self.assertEqual(len(data), 3)
        self.assertListEqual(data[0].pop('types'), [])
        self.assertListEqual(data[1].pop('types'), [])
        types = data[2].pop('types')
        for t in types:
            del t['id']
        self.assertCountEqual(types, EXPECTED_CORPUS_TYPES)

        self.assertListEqual(
            data,
            [
                {
                    'id': str(self.corpus_private.id),
                    'rights': ['read', 'write', 'admin'],
                    'public': False,
                    'indexable': False,
                    'name': 'B Private',
                    'description': '',
                    'created': str(FAKE_NOW).replace(' ', 'T') + 'Z',
                    'authorized_users': 2,
                    'thumbnail': None,
                    'thumbnail_url': None,
                    'top_level_type': None,
                },
                {
                    'id': str(self.corpus_hidden.id),
                    'rights': ['read', 'write', 'admin'],
                    'public': False,
                    'indexable': False,
                    'name': 'C Hidden',
                    'description': '',
                    'created': str(FAKE_NOW).replace(' ', 'T') + 'Z',
                    'authorized_users': 0,
                    'thumbnail': None,
                    'thumbnail_url': None,
                    'top_level_type': None,
                },
                {
                    'id': str(self.corpus_public.id),
                    'rights': ['read', 'write', 'admin'],
                    'public': True,
                    'indexable': False,
                    'name': 'Unit Tests',
                    'description': '',
                    'created': DB_CREATED,
                    'authorized_users': 1,
                    'thumbnail': None,
                    'thumbnail_url': None,
                    'top_level_type': None,
                }
            ]
        )

    def test_mixin(self):
        vol1 = Element.objects.get(name="Volume 1")
        vol2 = Element.objects.get(name="Volume 2")

        response = self.client.get(
            reverse('api:corpus-elements', kwargs={'corpus': self.corpus_private.id}),
            data={'type': 'volume'}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        response = self.client.get(
            reverse('api:corpus-elements', kwargs={'corpus': self.corpus_public.id}),
            data={'type': 'volume'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(len(data['results']), 2)
        self.assertCountEqual(
            [r['id'] for r in data['results']],
            [str(vol1.id), str(vol2.id)],
        )

    def test_create_requires_login(self):
        response = self.client.post(reverse('api:corpus'), {
            'name': 'New Corpus',
            'description': 'Some description',
            'public': False,
        })
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_public_admin(self):
        """
        Administrators can create a public corpus
        """
        self.client.force_login(self.superuser)
        response = self.client.post(reverse('api:corpus'), {
            'name': 'New Corpus',
            'description': 'Some description',
            'public': True,
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertTrue(Corpus.objects.get(name='New Corpus').public)

    def test_create_public_normal_user(self):
        """
        Normal users cannot create a public corpus
        """
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:corpus'), {
            'name': 'New Corpus',
            'description': 'Some description',
            'public': True,
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'public': ['Only admin users are allowed to edit the public attribute on a corpus.']
        })

    def test_create(self):
        """
        Any user is able to create a corpus defining its name and description
        """
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:corpus'), {
            'name': 'New Corpus',
            'description': 'Some description',
            'public': False,
        })
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        corpus = Corpus.objects.get(name='New Corpus', description='Some description', public=False)
        self.assertNotIn(self.corpus_private, Corpus.objects.admin(self.user))

        # Assert defaults types are set on the new corpus
        self.assertCountEqual(
            list(corpus.types.values(
                'slug',
                'display_name',
                'folder',
            )),
            [{
                'folder': False,
                **values
            } for values in DEFAULT_CORPUS_TYPES]
        )

    def test_retrieve_public(self):
        with self.assertNumQueries(3):
            response = self.client.get(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_public.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        types = data.pop('types')
        for t in types:
            del t['id']
        self.assertCountEqual(types, EXPECTED_CORPUS_TYPES)
        self.assertDictEqual(data, {
            'id': str(self.corpus_public.id),
            'name': self.corpus_public.name,
            'description': self.corpus_public.description,
            'public': True,
            'indexable': False,
            'rights': ['read'],
            'created': DB_CREATED,
            'authorized_users': 1,
            'thumbnail': None,
            'thumbnail_url': None,
            'top_level_type': None,
        })

    def test_retrieve(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'id': str(self.corpus_private.id),
            'name': self.corpus_private.name,
            'description': self.corpus_private.description,
            'public': False,
            'indexable': False,
            'rights': ['read', 'write'],
            'types': [],
            'created': str(FAKE_NOW).replace(' ', 'T') + 'Z',
            'authorized_users': 2,
            'thumbnail': None,
            'thumbnail_url': None,
            'top_level_type': None,
        })

    def test_retrieve_requires_login(self):
        response = self.client.get(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update(self):
        self.client.force_login(self.corpus_admin)
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}), {
            'name': 'new name',
            'description': 'new description',
        })
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.corpus_private.refresh_from_db()
        self.assertEqual(self.corpus_private.name, 'new name')
        self.assertEqual(self.corpus_private.description, 'new description')

    def test_update_thumbnail(self):
        self.client.force_login(self.corpus_admin)
        element = self.corpus_private.elements.create(
            type=self.corpus_private.types.create(
                slug='my_type',
                display_name='My type',
            ),
            image=Image.objects.first(),
            polygon=((0, 0), (0, 10), (10, 10), (10, 0), (0, 0)),
            name='An element',
        )
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}), {
            'thumbnail': str(element.id),
        })
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.corpus_private.refresh_from_db()
        self.assertEqual(self.corpus_private.thumbnail, element)

    def test_update_thumbnail_none(self):
        self.client.force_login(self.corpus_admin)
        element = self.corpus_private.elements.create(
            type=self.corpus_private.types.create(
                slug='my_type',
                display_name='My type',
            ),
            image=Image.objects.first(),
            polygon=((0, 0), (0, 10), (10, 10), (10, 0), (0, 0)),
            name='An element',
        )
        self.corpus_private.thumbnail = element
        self.corpus_private.save()
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}), {
            'thumbnail': None
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.corpus_private.refresh_from_db()
        self.assertIsNone(self.corpus_private.thumbnail)

    def test_update_thumbnail_no_zone(self):
        self.client.force_login(self.corpus_admin)
        folder = self.corpus_private.elements.create(
            type=self.corpus_private.types.create(slug='folder', folder=True),
            name='A folder',
        )
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}), {
            'thumbnail': str(folder.id)
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'thumbnail': [f'Invalid pk "{folder.id}" - object does not exist.']
        })

    def test_update_thumbnail_wrong_corpus(self):
        self.client.force_login(self.corpus_admin)
        element = self.corpus_public.elements.filter(polygon__isnull=False).first()
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}), {
            'thumbnail': str(element.id)
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'thumbnail': [f'Invalid pk "{element.id}" - object does not exist.']
        })

    def test_update_private_to_public_normal_user(self):
        """
        A normal user should not be able to make a private corpus public
        """
        self.client.force_login(self.corpus_admin)
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}), {
            'public': True
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'public': ['Only admin users are allowed to edit the public attribute on a corpus.']
        })

    def test_update_public_to_private_normal_user(self):
        """
        A normal user should not be able to make a public corpus private
        even if he has a write access to this corpus
        """
        self.client.force_login(self.user)
        self.assertIn(self.corpus_private, Corpus.objects.writable(self.user))
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_public.id}), {
            'public': False
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'public': ['Only admin users are allowed to edit the public attribute on a corpus.']
        })

    def test_normal_user_update_public_corpus(self):
        """
        An user with no write right should not be able to edit a public corpus
        """
        self.client.force_login(self.user)
        self.corpus_public.memberships.filter(user=self.user).update(level=Role.Guest.value)
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_public.id}), {
            'name': 'Gloubiboulga',
            'description': 'Bla bla bla'
        })
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have admin access to this corpus.'})

    def test_update_requires_login(self):
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}), {
            'name': 'new name',
            'description': 'new description',
        })
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_requires_admin_right_on_corpus(self):
        self.client.force_login(self.user)
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}), {
            'name': 'new name',
            'description': 'new description',
        })
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have admin access to this corpus.'})

    def test_update_top_level_type_same_corpus(self):
        """
        Cannot set a corpus' top-level type to a type from another corpus
        """
        self.client.force_login(self.superuser)
        self.corpus_private.types.create(slug='nope')
        self.assertFalse(self.corpus_public.types.filter(slug='nope').exists())
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_public.id}), {
            'top_level_type': 'nope',
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'top_level_type': ['Object with slug=nope does not exist.']})

    def test_update_top_level_type(self):
        self.client.force_login(self.superuser)
        self.assertIsNone(self.corpus_public.top_level_type)
        response = self.client.patch(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_public.id}), {
            'top_level_type': 'page',
        })
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.corpus_public.refresh_from_db()
        self.assertEqual(self.corpus_public.top_level_type.slug, 'page')

    def test_delete_requires_login(self):
        response = self.client.delete(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_requires_admin_right(self):
        self.assertNotIn(self.corpus_private, Corpus.objects.admin(self.user))
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:corpus-retrieve', kwargs={'pk': self.corpus_private.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch('arkindex.project.triggers.documents_tasks.corpus_delete.delay')
    def test_delete(self, delay_mock):
        self.client.force_login(self.user)
        self.corpus_private.memberships.filter(user=self.user).update(level=Role.Admin.value)
        response = self.client.delete(reverse('api:corpus-retrieve', kwargs={'pk': str(self.corpus_private.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=str(self.corpus_private.id),
            description=f'Deletion of corpus {self.corpus_private.name}',
            user_id=self.user.id,
        ))

    def test_delete_selected_elements_requires_login(self):
        response = self.client.delete(reverse('api:corpus-delete-selection', kwargs={'pk': self.corpus_private.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_selected_elements_requires_verified(self):
        # User is a corpus admin but is also unverified
        self.user.verified_email = False
        self.user.save()
        self.corpus_private.memberships.filter(user=self.user).update(level=Role.Admin.value)
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:corpus-delete-selection', kwargs={'pk': self.corpus_private.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_selected_elements_requires_admin_right(self):
        self.assertNotIn(self.corpus_private, Corpus.objects.admin(self.user))
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:corpus-delete-selection', kwargs={'pk': self.corpus_private.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_delete_selected_elements_inexistent_corpus_id(self):
        self.client.force_login(self.superuser)
        response = self.client.delete(reverse('api:corpus-delete-selection', kwargs={'pk': '12341234-1234-1234-1234-123412341234'}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch('arkindex.project.triggers.documents_tasks.element_trash.delay')
    def test_delete_selected_elements(self, delay_mock):
        act_type, _ = self.corpus_private.types.get_or_create(
            slug='act',
            display_name='Act',
        )
        act_x = Element.objects.create(
            type=act_type,
            corpus=self.corpus_private,
            name='Act xxx',
        )
        Element.objects.create(
            type=act_type,
            corpus=self.corpus_private,
            name='Act yyy',
        )
        self.superuser.selected_elements.add(act_x)

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.delete(reverse('api:corpus-delete-selection', kwargs={'pk': self.corpus_private.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        args, kwargs = delay_mock.call_args
        self.assertEqual(len(args), 0)
        self.assertCountEqual(list(kwargs.pop('queryset')), list(Element.objects.filter(id=act_x.id)))
        self.assertDictEqual(kwargs, {
            'delete_children': True,
            'user_id': self.superuser.id,
            'description': 'Element deletion',
        })
