import uuid
from datetime import datetime, timedelta, timezone

import sqlparse
from django.db.models.sql.constants import LOUTER
from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import Corpus, Element, MetaType
from arkindex.project.tests import FixtureAPITestCase


class TestListElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')

    def _get_prefetch_data(self, query):
        """
        Retrieve a set of UUIDs or None, and the string holding them in the query,
        from an SQL query string generated by Django's prefetching.
        This only handles models that have UUIDs as primary keys.

        This allows you to reuse it in assertExactQueries:

        with self.assertExactQueries('file.sql', params={…}) as ctx:
            expected_ids = set(...)
            query = next(query['sql'] for query in ctx if 'images_image' in query['sql'])
            actual_ids = self._get_prefetch_data(query)
            self.assertSetEqual(expected_ids, actual_ids)
        """
        # sqlparse.parse returns a tuple of statements, we will only have one
        parsed_query, = sqlparse.parse(query)

        # Get the where clause
        where = next(token for token in parsed_query.tokens if isinstance(token, sqlparse.sql.Where))
        # Get the IN comparison in the where clause
        comparison = next(token for token in where.tokens if isinstance(token, sqlparse.sql.Comparison))
        # Find the parentheses
        parens = next(token for token in comparison.tokens if isinstance(token, sqlparse.sql.Parenthesis))
        # Find the identifier list
        id_list = next(token for token in parens.tokens if isinstance(token, sqlparse.sql.IdentifierList))

        # Parse the identifier list, handling both UUIDs and NULL.
        ids = set()
        for id_token in id_list.tokens:
            # In our case, the Identifier is a string parsed as a UUID: it is made of three Tokens.
            # The first token is the UUID as a string, the second is the cast operator (::) and the third is the UUID type.
            # We grab the string, remove its quotes, and parse it as a Python UUID.
            if isinstance(id_token, sqlparse.sql.Identifier):
                ids.add(uuid.UUID(id_token.tokens[0].value.strip("'")))
            # NULL is parsed as a single Token class with a Keyword ttype.
            elif id_token.match(sqlparse.tokens.Keyword, "NULL"):
                ids.add(None)

        return ids

    def test_list_elements_sql(self):
        """
        Test ListElements with assertExactQueries.

        ListElements prefetches images, servers and element types in two separate queries.
        Django gathers image and type IDs from elements in an unordered manner, and performs the queries in a random
        order, which means that we cannot easily put the prefetch queries in the SQL fixture.

        To overcome this, this test uses sqlparse to look through the AST for the prefetch query and find the list of
        images and types, then builds a list of UUIDs and compares it against an expected list of IDs.
        It then removes the queries from the assertExactQueries context so that they are not included in the
        SQL fixture assertion.

        Future maintainers: If you want to dig into a prefetch query, the simplest way is to explore the result of
        sqlparse.parse in an interactive shell.  You could copy-paste an SQL string from the file in sql_validation,
        or run pdb straight from this test:

        image_query = sql_parse.parse(…)
        import pdb
        pdb.set_trace()

        Then use `p image_query` to see the query, `p image_query.tokens` to see its child tokens, etc.
        You can access an image ID directly using `p image_query.tokens[-1].tokens[-1].tokens[-1].tokens[1].tokens[0].value`.
        Learn more about sqlparse tokens here: https://sqlparse.readthedocs.io/en/latest/analyzing/
        """

        expected_elements = self.corpus.elements.all().order_by('name', 'id')[:20]
        # Since Django 3.2, NULL values are not prefetched since a primary key cannot be NULL.
        expected_image_ids = set(filter(None, expected_elements.values_list('image_id', flat=True)))
        expected_type_ids = set(expected_elements.values_list('type_id', flat=True))

        with self.assertExactQueries('list_elements.sql', params={'corpus_id': self.corpus.id}) as ctx:
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

            # Test the image prefetch separately to avoid the aforementioned issues
            image_index, image_query = next(
                (i, query['sql'])
                for i, query in enumerate(ctx)
                if 'images_image' in query['sql']
            )
            actual_image_ids = self._get_prefetch_data(image_query)
            self.assertSetEqual(expected_image_ids, actual_image_ids, 'Unexpected prefetched image IDs')

            # Django's prefetch queries can be in a random order; we just drop them from the captured queries as we
            # asserted on them manually, so that they can be removed from the SQL fixture entirely.
            # `ctx[i]` just forwards to `ctx.captured_queries[i]`, which is a property using `ctx.connection.queries`.
            # `queries` is a property on a Django connection that is defined as `list(self.queries_log)`, and
            # `queries_log` is a collections.deque instance.
            # To drop the queries, we need to delete from that deque directly.
            # As assertExactQueries might have been triggered after some queries were already logged by the Django
            # connection, it has an 'initial_queries' index which is used in .captured_queries, so we have to make
            # our indexes relative to that.
            del ctx.connection.queries_log[ctx.initial_queries + image_index]

            # Rinse and repeat for the type prefetch.
            type_index, type_query = next(
                (i, query['sql'])
                for i, query in enumerate(ctx)
                if 'documents_elementtype' in query['sql']
            )
            actual_type_ids = self._get_prefetch_data(type_query)
            self.assertSetEqual(expected_type_ids, actual_type_ids, 'Unexpected prefetched type IDs')
            del ctx.connection.queries_log[ctx.initial_queries + type_index]

    def test_list_elements_filter_name(self):
        nameSelected = 'ol'
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'name': nameSelected}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        for result in results:
            self.assertIn(nameSelected, result['name'])

    def test_list_elements_corpus_uuid4(self):
        """
        Ensure the corpus filter allows any UUID, not just version 4 UUIDs.
        Using the version=x argument on uuid.UUID causes the version and variant bits
        of a UUID to be changed on instanciation; since we generate v4 RFC 4122-compliant UUIDs,
        this would work in most cases - but we could have any UUID version or variant,
        generated by some other tool, which would cause strange 404 errors.
        """
        str_id = '12345678-1234-5678-1234-567812345678'
        self.assertNotEqual(
            uuid.UUID(str_id),
            uuid.UUID(str_id, version=4),
        )
        corpus = Corpus.objects.create(id=uuid.UUID(str_id), name='A corpus', public=True)
        element_type = corpus.types.create(slug='element', display_name='Generic element')
        element = corpus.elements.create(type=element_type, name='Something')

        response = self.client.get(reverse('api:corpus-elements', kwargs={'corpus': str_id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(element.id))
        self.assertEqual(data['results'][0]['corpus']['id'], str_id)

    def test_list_elements_filter_type(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'type': 'act'}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        for result in results:
            self.assertEqual(result['type'], 'act')

    def test_list_with_has_children(self):
        # Iterate over every page of results of this endpoint
        page = 1
        results = {}
        while True:
            with self.assertNumQueries(7):
                response = self.client.get(
                    reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                    data={'with_has_children': True, 'page': page},
                )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.json()
            for element in data['results']:
                results[element['name']] = element['has_children']
            if not data['next']:
                break
            page += 1

        self.assertDictEqual(
            results,
            {
                'Act 1': True,
                'Act 2': True,
                'Act 3': True,
                'Act 4': True,
                'Act 5': True,
                'Text line': False,
                'Surface A': False,
                'Surface B': False,
                'Surface C': False,
                'Surface D': False,
                'Surface E': False,
                'Surface F': False,
                'Volume 1': True,
                'Volume 1, page 1r': True,
                'Volume 1, page 1v': True,
                'Volume 1, page 2r': True,
                'Volume 2': True,
                'Volume 2, page 1r': False,
                'Volume 2, page 1v': False,
                'Volume 2, page 2r': False,
                'DATUM': False,
                'PARIS': False,
                'ROY': False,
            }
        )

    def test_list_top_level_left_join(self):
        """
        Ensure the top_level option on ListElements triggers a left join.
        """
        query = self.corpus.elements.filter(paths__isnull=True, name='something').query
        # Ensure the documents_elementpath table is joined via a LEFT JOIN
        self.assertEqual(query.alias_map['documents_elementpath'].join_type, LOUTER)
        # Ensure the IS NULL goes last
        self.assertTrue(str(query).endswith('AND "documents_elementpath"."id" IS NULL)'))

    def test_list_top_level(self):
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'top_level': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 1', 'Volume 2'],
        )

    def test_list_top_level_default_type(self):
        """
        Do not use the corpus top level type to filter top level elements when available;
        the frontend or API users are in charge of that.
        """
        # Add a top level element of type page
        element_type = self.corpus.types.create(slug='top_level', display_name='Some type')
        self.corpus.elements.create(type=element_type, name='Generic element')
        self.corpus.top_level_type = element_type
        self.corpus.save()

        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'top_level': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Generic element', 'Volume 1', 'Volume 2'],
        )

    def test_list_top_level_non_folder(self):
        """
        The top_level filter should allow non-folder elements
        """
        self.corpus.elements.create(type=self.corpus.types.filter(folder=False).first(), name='lonely page')

        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'top_level': True},
            )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['lonely page', 'Volume 1', 'Volume 2']
        )

    def test_list_non_top_level(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'top_level': False},
            )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            [
                'Act 1',
                'Act 2',
                'Act 3',
                'Act 4',
                'Act 5',
                'DATUM',
                'DATUM',
                'DATUM',
                'PARIS',
                'PARIS',
                'PARIS',
                'ROY',
                'ROY',
                'ROY',
                'Surface A',
                'Surface B',
                'Surface C',
                'Surface D',
                'Surface E',
                'Surface F',
            ]
        )

    def test_list_elements_filter_no_worker_version(self):
        self.corpus.elements.update(worker_version=self.worker_version)
        self.corpus.elements.filter(name__contains='Volume 2').update(worker_version=None)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'worker_version': False},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 2', 'Volume 2, page 1r', 'Volume 2, page 1v', 'Volume 2, page 2r'],
        )

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'worker_version': 0},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 2', 'Volume 2, page 1r', 'Volume 2, page 1v', 'Volume 2, page 2r'],
        )

    def test_list_elements_filter_worker_version(self):
        self.corpus.elements.filter(name__contains='Volume 2').update(worker_version=self.worker_version)

        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'worker_version': str(self.worker_version.id)},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 2', 'Volume 2, page 1r', 'Volume 2, page 1v', 'Volume 2, page 2r'],
        )

    def test_list_elements_filter_worker_version_validation(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'worker_version': 'blah'},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['Invalid UUID']})

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'worker_version': uuid.uuid4()},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['This worker version does not exist.']})

    def test_list_elements_filter_modified_since_not_modified(self):
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
        self.assertEqual(response.status_code, status.HTTP_304_NOT_MODIFIED)

    def test_list_elements_filter_modified_since_bad_format(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                HTTP_IF_MODIFIED_SINCE='AAAAAAAAAAAAAAAAA',
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'If-Modified-Since': ['Bad date format']})

    def test_list_elements_filter_modified_since(self):
        self.corpus.elements.filter(name='Volume 1').update(updated='2099-04-02T13:37:43Z')
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['count'], self.corpus.elements.count())

    def test_list_elements_filter_metadata_name(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'metadata_name': 'folio'}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            [
                'Volume 1, page 1r',
                'Volume 1, page 1v',
                'Volume 1, page 2r',
                'Volume 2, page 1r',
                'Volume 2, page 1v',
                'Volume 2, page 2r',
            ]
        )

    def test_list_elements_filter_metadata_value(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'metadata_value': '1v'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_value': ['This filter is not supported without metadata_name.'],
        })

    def test_list_elements_filter_metadata_name_value(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'metadata_name': 'folio', 'metadata_value': '1v'}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 1, page 1v', 'Volume 2, page 1v']
        )

    def test_list_elements_filter_metadata_operator(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'metadata_operator': 'gt'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_operator': ['This option is not supported without metadata_value.']
        })

    def test_list_elements_filter_metadata_operator_value(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'metadata_operator': 'gt', 'metadata_value': 4}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_value': ['This filter is not supported without metadata_name.'],
        })

    def test_list_elements_filter_metadata_operator_invalid(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'metadata_operator': 'aaaaaa'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_operator': [
                'This metadata operator is not supported.',
                'This option is not supported without metadata_value.',
            ]
        })

    def test_list_elements_filter_numeric_metadata_invalid_value(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'metadata_name': 'a', 'metadata_value': 'lol', 'metadata_operator': 'gt'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_value': ['Value must be a float when using a numeric comparison operator.'],
        })

    def test_list_elements_filter_numeric_metadata(self):
        element_type = self.corpus.types.first()
        for i in range(1, 6):
            element = self.corpus.elements.create(type=element_type, name=f'Element {i}')
            element.metadatas.create(name='number', type=MetaType.Numeric, value=str(i))
            element.metadatas.create(name='number', type=MetaType.Text, value='a')

        # operator, value, expected element names
        parameters = [
            ('lt', 2, ['Element 1']),
            ('lte', 2, ['Element 1', 'Element 2']),
            ('gt', 3, ['Element 4', 'Element 5']),
            ('gte', 3, ['Element 3', 'Element 4', 'Element 5']),
        ]

        for operator, value, expected_elements in parameters:
            with self.subTest(operator=operator, value=value):
                with self.assertNumQueries(4):
                    response = self.client.get(
                        reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                        data={
                            'metadata_name': 'number',
                            'metadata_value': value,
                            'metadata_operator': operator,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    expected_elements
                )

    def test_list_elements_filter_metadata_contains(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={
                    'metadata_name': 'folio',
                    'metadata_operator': 'contains',
                    'metadata_value': 'v',
                }
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 1, page 1v', 'Volume 2, page 1v']
        )

    def test_list_elements_with_corpus_false(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id})
                + '?with_corpus=false'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertSetEqual(
            set(element.get('corpus') for element in response.json()['results']),
            {None}
        )

    def test_list_elements_page_size(self):
        """
        Corpus elements can be listed with a large page size
        """
        corpus = Corpus.objects.create(name="A large corpus", public=True)
        folder_type = corpus.types.create(slug='folder', display_name='Folder')
        # Create 99 folder on this corpus
        Element.objects.bulk_create(
            Element(corpus=corpus, name=f"folder {i}", type=folder_type)
            for i in range(1, 100)
        )
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': corpus.id}),
                {'page_size': 50}
            )
        data = response.json()
        self.assertEqual(len(data['results']), 50)
        self.assertEqual(data['count'], 99)

    def test_list_elements_page_size_limit(self):
        """
        Page size is limited to 500 items
        """
        corpus = Corpus.objects.create(name="A large corpus", public=True)
        folder_type = corpus.types.create(slug='folder', display_name='Folder')
        # Create 99 folder on this corpus
        Element.objects.bulk_create(
            Element(corpus=corpus, name=f"folder {i}", type=folder_type)
            for i in range(1, 1000)
        )
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': corpus.id}),
                {'page_size': 800}
            )
        data = response.json()
        self.assertEqual(len(data['results']), 500)
        self.assertEqual(data['count'], 999)

    def test_list_elements_invalid_sort(self):
        cases = [
            ({'order': 'blah', 'order_direction': 'asc'}, {'order': ['Unknown sorting field']}),
            ({'order': 'name', 'order_direction': 'left'}, {'order_direction': ['Unknown sorting direction']}),
            ({'order': 'blah', 'order_direction': 'left'}, {'order': ['Unknown sorting field']}),
        ]
        for params, expected in cases:
            with self.subTest(**params):
                response = self.client.get(
                    reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                    params,
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected)

    def test_list_elements_sort_created(self):
        elements = list(self.corpus.elements.order_by('?').only('id'))
        created = datetime(2020, 2, 2, tzinfo=timezone.utc)
        for element in elements:
            element.created = created
            created += timedelta(seconds=1)
        Element.objects.bulk_update(elements, ['created'])

        cases = [
            ('asc', elements[:20]),
            ('desc', reversed(elements[-20:])),
        ]
        for direction, expected in cases:
            with self.subTest(direction=direction), self.assertNumQueries(5):
                response = self.client.get(
                    reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                    {'order': 'created', 'order_direction': direction}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['id'] for element in response.json()['results']],
                    [str(element.id) for element in expected]
                )
