from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import Corpus, Element
from arkindex.images.models import ImageServer
from arkindex.project.aws import S3FileStatus
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestCreateElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume_type = cls.corpus.types.get(slug='volume')
        cls.page_type = cls.corpus.types.get(slug='page')
        cls.act_type = cls.corpus.types.get(slug='act')
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.element = Element.objects.get(name='Volume 1, page 2r')
        cls.image = ImageServer.objects.local.images.create(
            path="kingdom/far/away",
            status=S3FileStatus.Checked,
            width=42,
            height=42,
        )
        cls.worker_version = WorkerVersion.objects.get(worker__slug='dla')

        # The image is always the same in API responses
        cls.image_response = {
            'id': str(cls.image.id),
            's3_url': None,
            'width': 42,
            'height': 42,
            'path': 'kingdom/far/away',
            'url': 'http://server/kingdom/far/away',
            'status': 'checked',
            'server': {
                'display_name': 'Test Server',
                'max_height': None,
                'max_width': None,
                'url': 'http://server',
            }
        }

    def make_create_request(self, name='default', corpus=None, elt_type='volume', **options):
        request = {
            'path': reverse('api:elements-create'),
            'data': {
                'type': elt_type,
                'name': name,
                'corpus': corpus or str(self.corpus.id),
            },
            'format': 'json',
        }
        # Optional parameters
        request['data'].update(options)
        return request

    def test_create_element_not_verified(self):
        self.assertEqual(
            self.client.post(**self.make_create_request('no')).status_code,
            status.HTTP_403_FORBIDDEN
        )

    def test_create_element_volume(self):
        # Create a Volume
        self.client.force_login(self.user)
        request = self.make_create_request('my new volume')
        with self.assertNumQueries(9):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())
        data = response.json()
        self.assertIn('id', data)
        volume = Element.objects.get(id=data['id'])
        self.assertDictEqual(
            data,
            {
                'id': str(volume.id),
                'type': volume.type.slug,
                'name': volume.name,
                'thumbnail_put_url': None,
                'thumbnail_url': volume.thumbnail.s3_url,
                'worker_version': None,
                'creator': 'Test user',
                'rights': ['read', 'write', 'admin'],
                'classifications': [],
                'corpus': {
                    'id': str(volume.corpus.id),
                    'name': volume.corpus.name,
                    'public': volume.corpus.public
                },
                'zone': None,
                'rotation_angle': 0,
                'mirrored': False,
                'created': volume.created.isoformat().replace('+00:00', 'Z'),
            }
        )
        self.assertEqual(volume.name, 'my new volume')
        self.assertEqual(volume.type, self.volume_type)
        self.assertEqual(volume.creator, self.user)

    def test_create_volume_page(self):
        # Add a page with an Image to an existing volume
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            name='The castle of my dreams',
            image=str(self.image.id),
        )
        with self.assertNumQueries(16):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        page = self.corpus.elements.get(id=data['id'])
        self.assertDictEqual(
            data,
            {
                'id': str(page.id),
                'name': page.name,
                'type': page.type.slug,
                'thumbnail_put_url': None,
                'thumbnail_url': None,
                'created': page.created.isoformat().replace('+00:00', 'Z'),
                'creator': 'Test user',
                'rights': ['read', 'write', 'admin'],
                'worker_version': None,
                'classifications': [],
                'corpus': {
                    'id': str(page.corpus.id),
                    'name': page.corpus.name,
                    'public': page.corpus.public
                },
                'zone': {
                    'id': str(page.id),
                    'image': self.image_response,
                    'polygon': [[0, 0], [0, 42], [42, 42], [42, 0], [0, 0]],
                    'url': page.iiif_url,
                },
                'rotation_angle': 0,
                'mirrored': False,
            }
        )
        self.assertEqual(page.name, 'The castle of my dreams')
        self.assertEqual(page.type, self.page_type)
        self.assertEqual(page.creator, self.user)

    def test_create_inherit_parent_image(self):
        """
        Creating an element with a parent, a polygon and no image should inherit the parent's image
        """
        polygon = [[10, 10], [10, 40], [40, 40], [40, 10], [10, 10]]
        self.vol.image = self.image
        self.vol.polygon = polygon
        self.vol.save()
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            name='The castle of my dreams',
            polygon=polygon,
        )
        with self.assertNumQueries(16):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        page = self.corpus.elements.get(id=data['id'])
        self.assertEqual(response.json()['zone'], {
            'id': str(page.id),
            'image': self.image_response,
            'polygon': polygon,
            'url': page.iiif_url,
        })

    def test_create_volume_act(self):
        # Create an volume child of type Act
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            name='Castle story',
            elt_type='act'
        )
        with self.assertNumQueries(15):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        act = Element.objects.get(id=response.json()['id'])
        self.assertEqual(act.name, 'Castle story')
        self.assertEqual(act.type, self.act_type)
        self.assertEqual(act.creator, self.user)

    def test_create_element_worker_version(self):
        # Create an element with a worker version
        self.client.force_login(self.internal_user)
        request = self.make_create_request(
            name='Castle story',
            elt_type='act',
            worker_version=str(self.worker_version.id),
        )

        with self.assertNumQueries(7):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        act = Element.objects.get(id=response.json()['id'])
        self.assertEqual(act.name, 'Castle story')
        self.assertEqual(act.type, self.act_type)
        self.assertEqual(act.worker_version, self.worker_version)
        self.assertIsNone(act.creator)

    def test_create_element_worker_version_requires_internal(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            name='Castle story',
            elt_type='act',
            worker_version=str(self.worker_version.id),
        )

        with self.assertNumQueries(7):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            'worker_version': ['Only an internal user can create an element with a worker version.'],
        })

    def test_create_element_polygon_no_image(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            elt_type='page',
            name='The castle of my dreams again',
            polygon=[[10, 10], [10, 40], [40, 40], [40, 10], [10, 10]],
        )
        with self.assertNumQueries(6):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'polygon': ['An image or a parent with an image is required to create an element with a polygon.'],
        })

    def test_create_element_polygon_negative(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            name='The castle of my dreams again',
            image=str(self.image.id),
            polygon=[[0, 0], [10, 10], [40, 40], [-10, 10], [0, 0]]
        )
        with self.assertNumQueries(8):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'polygon': {'3': {'0': ['Ensure this value is greater than or equal to 0.']}}
        })

    def test_create_element_polygon(self):
        # Create an element with a polygon to an existing volume
        polygon = [[10, 10], [10, 40], [40, 40], [40, 10], [10, 10]]
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            name='The castle of my dreams again',
            image=str(self.image.id),
            polygon=polygon
        )
        with self.assertNumQueries(16):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        page = self.corpus.elements.get(id=data['id'])
        self.assertDictEqual(
            data,
            {
                'id': str(page.id),
                'name': page.name,
                'type': page.type.slug,
                'thumbnail_put_url': None,
                'thumbnail_url': None,
                'worker_version': None,
                'creator': 'Test user',
                'rights': ['read', 'write', 'admin'],
                'classifications': [],
                'corpus': {
                    'id': str(page.corpus.id),
                    'name': page.corpus.name,
                    'public': page.corpus.public
                },
                'zone': {
                    'id': str(page.id),
                    'polygon': polygon,
                    'image': self.image_response,
                    'url': page.iiif_url,
                },
                'rotation_angle': 0,
                'mirrored': False,
                'created': page.created.isoformat().replace('+00:00', 'Z'),
            }
        )
        self.assertEqual(page.name, 'The castle of my dreams again')
        self.assertEqual(page.type, self.page_type)

    def test_create_element_huge_polygon(self):
        # Create a polygon with points in a reversed order, because this is what ST_Normalize would return
        # (this makes assertions easier than handling a reordering)
        polygon = [[0, 0]] + [[i, i] for i in range(3000, -1, -1)]
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            name='The castle of my dreams again',
            image=str(self.image.id),
            polygon=polygon
        )
        with self.assertNumQueries(16):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())
        self.assertListEqual(response.json()['zone']['polygon'], polygon)

    def test_create_parent_different_corpus(self):
        self.client.force_login(self.user)
        new_corpus = Corpus.objects.create(name='new')
        new_corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        new_corpus.types.create(slug='act')
        request = self.make_create_request(
            parent=str(self.vol.id),
            corpus=str(new_corpus.id),
            elt_type='act'
        )
        with self.assertNumQueries(6):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'corpus': ['Parent and child must be in the same corpus']
        })

    def test_create_filter_type(self):
        """
        Ensure types are filtered by corpus
        """
        self.assertTrue(self.corpus.types.filter(slug='volume').exists())
        self.client.force_login(self.user)
        new_corpus = Corpus.objects.create(name='new')
        new_corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        new_vol_type = new_corpus.types.create(slug='volume', display_name='Volume', folder=True)
        request = self.make_create_request(
            corpus=str(new_corpus.id),
            elt_type='volume',
            name='something',
        )
        with self.assertNumQueries(8):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_vol = new_corpus.elements.get()
        self.assertEqual(new_vol.type, new_vol_type)
        self.assertEqual(new_vol.type.corpus, new_corpus)

    def test_create_wrong_type(self):
        self.assertFalse(self.corpus.types.filter(slug='kartoffelsalad').exists())
        self.client.force_login(self.user)
        request = self.make_create_request(
            corpus=str(self.corpus.id),
            elt_type='kartoffelsalad',
            name='something',
        )
        with self.assertNumQueries(6):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'type': ["Type 'kartoffelsalad' not found in corpus Unit Tests"],
        })

    def test_create_image_not_checked(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            image=str(self.image.id)
        )
        self.image.status = S3FileStatus.Error
        self.image.save()
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        element = self.corpus.elements.get(id=response.json()['id'])
        self.assertEqual(element.image.status, S3FileStatus.Error)

    def test_create_image_zero_dimensions(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            image=str(self.image.id)
        )
        self.image.width = 0
        self.image.height = 0
        self.image.status = S3FileStatus.Unchecked
        self.image.save()
        with self.assertNumQueries(8):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'image': ['This image does not have valid dimensions.']})

    def test_create_element_slim_output_simple(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            name='slim output !',
        )
        request['path'] += '?slim_output=true'
        with self.assertNumQueries(7):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        element = self.corpus.elements.get(id=response.json()['id'])
        self.assertDictEqual(response.json(), {
            "id": str(element.id),
        })

    def test_create_element_slim_output(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            name='slim output !',
            parent=str(self.vol.id),
            elt_type='page',
            image=str(self.image.id),
        )
        request['path'] += '?slim_output=true'
        with self.assertNumQueries(14):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        element = self.corpus.elements.get(id=response.json()['id'])
        self.assertDictEqual(response.json(), {
            "id": str(element.id),
        })

    def test_create_element_rotation_mirroring(self):
        self.client.force_login(self.user)

        self.vol.rotation_angle = 180
        self.vol.mirrored = True
        self.vol.save()

        # Rotation and mirroring sent in the request, and returned after inheritance from the volume
        parameters = [
            (0, False, 0, False),
            (None, False, 180, False),
            (None, None, 180, True),
            (0, None, 0, True),
        ]

        for rotation_angle, mirrored, expected_rotation_angle, expected_mirrored in parameters:
            with self.subTest(rotation_angle=rotation_angle, mirrored=mirrored):
                request = self.make_create_request(
                    parent=str(self.vol.id),
                    elt_type='page',
                    name='The castle of my dreams again',
                    image=str(self.image.id),
                )

                if rotation_angle is not None:
                    request['data']['rotation_angle'] = rotation_angle
                if mirrored is not None:
                    request['data']['mirrored'] = mirrored

                with self.assertNumQueries(16):
                    response = self.client.post(**request)
                    self.assertEqual(response.status_code, status.HTTP_201_CREATED)

                data = response.json()
                self.assertEqual(data['rotation_angle'], expected_rotation_angle)
                self.assertEqual(data['mirrored'], expected_mirrored)

    def test_create_element_rotation_validation(self):
        self.client.force_login(self.user)

        parameters = [
            (-1, 'Ensure this value is greater than or equal to 0.'),
            (360, 'Ensure this value is less than or equal to 359.'),
            (1.3, 'A valid integer is required.'),
        ]

        for rotation_angle, expected_message in parameters:
            with self.subTest(rotation_angle=rotation_angle):
                request = self.make_create_request(
                    parent=str(self.vol.id),
                    elt_type='page',
                    name='The castle of my dreams again',
                    image=str(self.image.id),
                    rotation_angle=rotation_angle,
                )

                with self.assertNumQueries(8):
                    response = self.client.post(**request)
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(response.json(), {
                    'rotation_angle': [expected_message]
                })
