from datetime import datetime, timedelta, timezone
from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import CorpusExportState
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestExport(FixtureAPITestCase):

    @patch('arkindex.project.triggers.export.export_corpus.delay')
    def test_start(self, delay_mock):
        self.client.force_login(self.superuser)
        response = self.client.post(reverse('api:corpus-export', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        export = self.corpus.exports.get()
        self.assertDictEqual(response.json(), {
            "id": str(export.id),
            "created": export.created.isoformat().replace('+00:00', 'Z'),
            "updated": export.updated.isoformat().replace('+00:00', 'Z'),
            "user": {
                "id": self.superuser.id,
                "display_name": self.superuser.display_name,
                "email": self.superuser.email,
            },
            "corpus_id": str(self.corpus.id),
            "state": CorpusExportState.Created.value,
        })

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_export=export,
            user_id=self.superuser.id,
            description='Export of corpus Unit Tests'
        ))

    @patch('arkindex.project.triggers.export.export_corpus.delay')
    def test_start_requires_login(self, delay_mock):
        response = self.client.post(reverse('api:corpus-export', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertFalse(self.corpus.exports.exists())
        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.export.export_corpus.delay')
    def test_start_requires_verified(self, delay_mock):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        response = self.client.post(reverse('api:corpus-export', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertFalse(self.corpus.exports.exists())
        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.export.export_corpus.delay')
    def test_start_requires_contributor(self, delay_mock):
        self.user.rights.update(level=Role.Guest.value)
        self.client.force_login(self.user)

        response = self.client.post(reverse('api:corpus-export', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertFalse(self.corpus.exports.exists())
        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.export.export_corpus.delay')
    def test_start_running(self, delay_mock):
        self.client.force_login(self.superuser)
        self.corpus.exports.create(user=self.user, state=CorpusExportState.Running)

        response = self.client.post(reverse('api:corpus-export', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ['An export is already running for this corpus.'])

        self.assertEqual(self.corpus.exports.count(), 1)
        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.export.export_corpus.delay')
    def test_start_recent_export(self, delay_mock):
        self.client.force_login(self.superuser)
        with patch('django.utils.timezone.now') as mock_now:
            mock_now.return_value = datetime.now(timezone.utc) - timedelta(hours=2)
            self.corpus.exports.create(
                user=self.user,
                state=CorpusExportState.Done,
            )

        response = self.client.post(reverse('api:corpus-export', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ['An export has already been made for this corpus in the last 6 hours.'])

        self.assertEqual(self.corpus.exports.count(), 1)
        self.assertFalse(delay_mock.called)

    def test_list(self):
        export1 = self.corpus.exports.create(user=self.user, state=CorpusExportState.Done)
        export2 = self.corpus.exports.create(user=self.superuser)
        self.client.force_login(self.superuser)

        response = self.client.get(reverse('api:corpus-export', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [
            {
                "id": str(export2.id),
                "created": export2.created.isoformat().replace('+00:00', 'Z'),
                "updated": export2.updated.isoformat().replace('+00:00', 'Z'),
                "state": CorpusExportState.Created.value,
                "user": {
                    "id": self.superuser.id,
                    "display_name": self.superuser.display_name,
                    "email": self.superuser.email,
                },
                "corpus_id": str(self.corpus.id),
            },
            {
                "id": str(export1.id),
                "created": export1.created.isoformat().replace('+00:00', 'Z'),
                "updated": export1.updated.isoformat().replace('+00:00', 'Z'),
                "state": CorpusExportState.Done.value,
                "user": {
                    "id": self.user.id,
                    "display_name": self.user.display_name,
                    "email": self.user.email,
                },
                "corpus_id": str(self.corpus.id),
            },
        ])

    def test_list_requires_login(self):
        response = self.client.get(reverse('api:corpus-export', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:corpus-export', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_requires_guest(self):
        self.user.rights.all().delete()
        self.corpus.public = False
        self.corpus.save()
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:corpus-export', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch('arkindex.project.aws.s3.meta.client.generate_presigned_url')
    def test_download_export(self, presigned_url_mock):
        presigned_url_mock.return_value = 'http://somewhere'
        self.client.force_login(self.superuser)
        export = self.corpus.exports.create(user=self.user, state=CorpusExportState.Done)
        response = self.client.get(reverse('api:download-export', kwargs={'pk': export.id}))
        self.assertEqual(response.status_code, status.HTTP_302_FOUND)
        self.assertEqual(response.headers['Location'], 'http://somewhere')

    def test_download_export_requires_login(self):
        export = self.corpus.exports.create(user=self.user, state=CorpusExportState.Done)
        response = self.client.get(reverse('api:download-export', kwargs={'pk': export.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_download_export_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        export = self.corpus.exports.create(user=self.user, state=CorpusExportState.Done)
        response = self.client.get(reverse('api:download-export', kwargs={'pk': export.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_download_export_requires_guest(self):
        self.user.rights.all().delete()
        self.corpus.public = False
        self.corpus.save()
        self.client.force_login(self.user)
        export = self.corpus.exports.create(user=self.user, state=CorpusExportState.Done)
        response = self.client.get(reverse('api:download-export', kwargs={'pk': export.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_download_export_not_done(self):
        self.client.force_login(self.superuser)
        for state in (CorpusExportState.Created, CorpusExportState.Running, CorpusExportState.Failed):
            export = self.corpus.exports.create(user=self.user, state=state)
            response = self.client.get(reverse('api:download-export', kwargs={'pk': export.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
