import logging

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from arkindex.images.models import Image, ImageServer
from arkindex.project.argparse import CorpusArgument, ElementArgument
from arkindex.project.aws import S3FileStatus

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s [%(levelname)s] %(message)s',
)
logger = logging.getLogger()


class Command(BaseCommand):
    help = 'Check images availability'

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            '--force',
            action='store_true',
            default=False,
        )
        parser.add_argument(
            '--corpus',
            help='A corpus to run checks on',
            type=CorpusArgument(),
        )
        parser.add_argument(
            '--element',
            help='One or more elements to run checks on',
            type=ElementArgument(),
            nargs='+',
        )
        parser.add_argument(
            '--sample',
            help='Also test a few already checked images from each server',
            nargs='?',
            const=settings.CHECK_IMAGES_SAMPLE_SIZE,
            default=0,
        )

    def handle(self, corpus=None, element=None, force=False, sample=0, **options):
        if corpus and element:
            raise CommandError('--corpus and --element cannot be used together')
        if (corpus or element) and sample:
            raise CommandError('--corpus and --element cannot be used with --sample')
        if force and sample:
            raise CommandError('--force and --sample cannot be used together')

        if corpus:
            images = Image.objects.filter(elements__corpus=corpus).distinct()
        elif element:
            images = Image.objects.filter(elements__in=element).distinct()
        else:
            images = Image.objects.all()

        if not force:
            images = images.exclude(status=S3FileStatus.Checked)

        if sample >= 1:
            # Re-check a few images from each server
            for server in ImageServer.objects.all():
                server_sample = server.images \
                                      .filter(status=S3FileStatus.Checked) \
                                      .order_by('?')[:sample]
                logger.info('Re-checking {} images in server {}'.format(len(server_sample), server.display_name))
                self.check(server_sample)
        self.check(images)

    def check(self, images=[]):
        successful, failed = 0, 0

        for image in images:
            logger.info('Checking image {} at {}'.format(str(image.id), image.url))
            image.perform_check(save=True)
            if image.status == S3FileStatus.Checked:
                successful += 1
            elif image.status == S3FileStatus.Error:
                failed += 1

        return successful, failed
