import logging

from rest_framework.exceptions import PermissionDenied
from rest_framework.permissions import SAFE_METHODS, BasePermission

logger = logging.getLogger(__name__)


def require_auth(request, view):
    return request.user.is_authenticated


def require_internal(request, view):
    return request.user.is_authenticated and request.user.is_internal


def require_verified_email(request, view):
    return request.user.is_authenticated and (request.user.is_admin or request.user.verified_email)


def _get_scopes(request, view):
    specific_scopes_attr = '{}_scopes'.format(request.method.lower())
    scopes = list(getattr(view, 'scopes', []))
    scopes.extend(getattr(view, specific_scopes_attr, []))
    return set(scopes)


def check_scopes(request, view):
    """
    When the `scopes` attribute is defined on the view, ensure an authenticated user has said scopes,
    no matter which HTTP method is used.
    When the `<method>_scopes` attribute is defined for a given HTTP method, ensures the user
    has said scopes only for this HTTP method.
    """
    missing_scopes = _get_scopes(request, view)
    # Skip this check if there are no defined scopes, if the user is not logged in, is an admin or is internal
    if not missing_scopes or not request.user.is_authenticated or request.user.is_admin or request.user.is_internal:
        return True

    if request.user.is_authenticated:
        # Tell which scopes are missing by subtracting found scopes from the expected scopes
        missing_scopes -= set(request.user.user_scopes.filter(scope__in=missing_scopes).values_list('scope', flat=True))

    if missing_scopes:
        # If scopes are not matched, cause a custom error
        raise PermissionDenied({
            'detail': PermissionDenied.default_detail,
            'missing_scopes': [scope.value for scope in missing_scopes],
        })

    return True


class ArkindexPermission(BasePermission):
    """
    An easier to extend permission class.

    You can add a list of functions to the `checks` class attribute; each function should return True or False
    depending on whether it allows or denies access.  It is also possible to raise custom exceptions to override
    other checks and get finer-grained control on the API response.

    `allow_safe_methods` is equivalent to the -IsReadOnly series of DRF's original permissions, where access is
    automatically granted without even running the checks if the user uses HEAD, GET or OPTIONS.
    """
    checks = ()
    allow_safe_methods = False

    def has_permission(self, request, view):
        if self.allow_safe_methods and request.method in SAFE_METHODS:
            return True

        for check in self.checks:
            if not check(request, view):
                logger.warning(f'Permission check {check.__name__} failed for user {request.user}')
                return False

        return True


class IsAuthenticated(ArkindexPermission):
    checks = (require_auth, check_scopes)


class IsAuthenticatedOrReadOnly(IsAuthenticated):
    allow_safe_methods = True


class IsInternal(ArkindexPermission):
    checks = (require_internal, check_scopes)


class IsVerified(ArkindexPermission):
    checks = (require_verified_email, check_scopes)


class IsVerifiedOrReadOnly(IsVerified):
    allow_safe_methods = True
