import re
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, Iterable, Mapping, Optional, Union

import sqlparse
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db import DEFAULT_DB_ALIAS, connections
from django.test import TestCase
from django.test.utils import CaptureQueriesContext
from django.utils.functional import cached_property
from rest_framework.test import APITestCase

from arkindex.documents.models import Corpus
from arkindex.images.models import ImageServer
from arkindex.users.models import Group, User


class _AssertExactQueriesContext(CaptureQueriesContext):
    """
    Context manager for assertExactQueries.
    The implementation is inspired by assertNumQueries's own implementation.
    """

    def __init__(self, test_case, path, params, connection, skip=0):
        self.test_case = test_case
        self.path = settings.SQL_VALIDATION_DIR / Path(path)
        self.params = params
        self.skip = skip
        super().__init__(connection)

    def _format(self, query: str) -> str:
        """
        Format one or more SQL queries.

        Does not allow more than two newlines at once (at most one empty line).
        This still allows separating with empty lines in the expected SQL file,
        but prevents duplicate empty lines from appearing;
        sqlparse.format duplicates an empty line when it encounters one, causing assertion errors.
        """
        return re.sub(
            r'\n\n+',
            '\n\n',
            sqlparse.format(
                query,
                reindent=True,
                strip_comments=True,
                use_space_around_operators=True,
                indent_width=4,
            ).strip()
        )

    def __exit__(self, exc_type, exc_value, traceback):
        super().__exit__(exc_type, exc_value, traceback)
        if exc_type is not None:
            return

        # Django's logged queries are each on a single line without semicolons
        # so we lint them and assemble them to build a more readable diff for humans.
        actual_sql = self._format(
            ';'.join(query['sql'] for query in self[self.skip:])
        )

        # When the file does not exist, try to create it with the current queries.
        # This avoids having to write all of the queries by hand when adding assertExactQueries for the first time.
        if not self.path.exists():
            try:
                self.path.write_text(actual_sql)
            except IOError as e:
                raise AssertionError(
                    f'Could not assert on SQL queries; the file at {self.path} did not exist. '
                    'A new file could not be created with the current SQL queries.'
                ) from e
            else:
                raise AssertionError(
                    f'Could not assert on SQL queries; the file at {self.path} did not exist. '
                    'A file has been created with the current SQL queries, please check it manually.'
                )

        # Ignore newlines in the expected SQL, otherwise the reformatting might have duplicate line breaks
        expected_sql = self.path.read_text().strip()

        # Apply str.format on the SQL when needed to handle variable IDs
        if self.params:
            if isinstance(self.params, Mapping):
                expected_sql = expected_sql.format(**self.params)
            else:
                expected_sql = expected_sql.format(*self.params)

        expected_sql = self._format(expected_sql)

        # Use the base TestCase's assertions, since they can already handle displaying a diff.
        self.test_case.assertEqual(expected_sql, actual_sql)


class FixtureMixin(object):
    """
    Add the database fixture to a test case
    """

    fixtures = ['data.json', ]

    @classmethod
    def setUpTestData(cls):
        cls.corpus = Corpus.objects.get(name='Unit Tests')
        cls.user = User.objects.get(email='user@user.fr')
        cls.group = Group.objects.get(name='User group')
        cls.superuser = User.objects.get(email='root@root.fr')
        cls.imgsrv = ImageServer.objects.get(url='http://server')
        # Clear content type cache after setting up test data in case a single test is run (see tearDown method above)
        ContentType.objects.clear_cache()

    @cached_property
    def internal_user(self):
        return User.objects.get(email='internal@internal.fr')

    def tearDown(self):
        # Clean content type cache for SQL requests checks consistency
        ContentType.objects.clear_cache()
        # Force clean the default farm global variable in `arkindex.dataimport.utils` module
        # This is required not to alter query counts and avoid caching a farm that does not exist in the fixture
        from arkindex.dataimport import utils
        setattr(utils, '__default_farm_id', None)

    @contextmanager
    def subTest(self, *args, **kwargs):
        with super().subTest(*args, **kwargs):
            yield
        # Clear content type cache for consistent SQL queries during sub-tests
        ContentType.objects.clear_cache()

    def assertExactQueries(
            self,
            path: Union[Path, str],
            func: Optional[Callable] = None,
            *args,
            using: str = DEFAULT_DB_ALIAS,
            params: Union[Iterable, Mapping[str, Any]] = [],
            skip: int = 0,
            **kwargs) -> Optional[_AssertExactQueriesContext]:
        """
        Assert that a function call causes exactly the SQL queries specified in a given file.
        When used as a context manager, asserts than any queries occurring while in the context match the given file.
        This function supports all of Django's `assertNumQueries` syntax:

        >>> self.assertExactQueries('path/to/file.sql', do_something)
        >>> with self.assertExactQueries('path/to/file.sql'):
        ...     do_something()

        Paths will be relative to the SQL_VALIDATION_DIR setting.
        SQL queries can often hold IDs or other dynamic values that depend on the test context.
        To avoid having to update IDs or rewrite the fixture file every time, you can use the `params` argument
        and give it anything supported by `str.format`; either an iterable or a mapping. For example:

        ```
        SELECT "documents_element"."name" FROM "documents_element" WHERE "documents_element"."id" = '{id}'::uuid;
        ```

        >>> with self.assertExactQueries('path/to/file.sql', params={'id': element_id}):
        ...     Element.objects.filter(id=element_id).values('name')

        To make first runs easier, when the expect SQL file does not exist, assertExactQueries will try to write
        the current SQL queries to it and warn you using an `AssertionError`. Make sure to check this new SQL file
        as subsequent runs of unit tests will no longer fail.

        In API endpoint tests, some queries do not really need to be tested and are annoying to mock, such as
        session queries that use the `django_session` table and are the first query to be made when you call an endpoint
        using the test client. You can skip those queries and go straight to the point using `skip`.
        This will skip the first query it finds:

        >>> with self.assertExactQueries('my_endpoint.sql', skip=1):
        ...     self.client.get(reverse('api:my-endpoint'))
        """
        context = _AssertExactQueriesContext(self, path, params, connections[using], skip=skip)

        if func is None:
            return context

        with context:
            func(*args, **kwargs)


class FixtureTestCase(FixtureMixin, TestCase):
    """
    Django test case with the database fixture
    """


class FixtureAPITestCase(FixtureMixin, APITestCase):
    """
    Django REST Framework test case with the database fixture
    """
