import subprocess
from pathlib import Path
from subprocess import CalledProcessError
from unittest.mock import call, patch

from django.conf import settings
from django.core.checks import Error, Warning
from django.test import TestCase, override_settings
from django.urls import path


@override_settings(ALL_CHECKS=True)
class ChecksTestCase(TestCase):

    def test_api_urls_check(self):
        '''
        Check should always be OK
        '''
        from arkindex.project.checks import api_urls_check
        from arkindex.project.api_v1 import api

        self.assertListEqual(api_urls_check(), [])

        # Add an invalid url
        bad_url = path('bad', lambda: None, name='bad-url')
        api.append(bad_url)

        # Should trigger the check
        self.assertListEqual(
            api_urls_check(),
            [
                Error(
                    'This api url misses a mandatory ending /',
                    hint=str(bad_url),
                    id='arkindex.E001',
                )
            ]
        )

    def test_local_imageserver_check(self):
        """
        Test the local imageserver existence check
        """
        from arkindex.project.checks import local_imageserver_check
        from arkindex.images.models import ImageServer
        self.assertEqual(ImageServer.objects.count(), 0)

        with self.settings(LOCAL_IMAGESERVER_ID=42):
            self.assertListEqual(local_imageserver_check(), [
                Error(
                    'Local ImageServer with ID 42 does not exist',
                    hint='settings.LOCAL_IMAGESERVER_ID = 42',
                    id='arkindex.E004',
                ),
            ])

        srv = ImageServer.objects.create(display_name="Test Local Server", url="http://nowhere")

        with self.settings(LOCAL_IMAGESERVER_ID=srv.id):
            self.assertListEqual(local_imageserver_check(), [])

        srv.delete()

    @patch('arkindex.project.checks.subprocess.run')
    @override_settings(
        ARKINDEX_TASKS_IMAGE='nuh',
    )
    def test_docker_images_check(self, run_mock):
        from arkindex.project.checks import docker_images_check

        expected_calls = [
            call(
                ['docker', 'image', 'inspect', 'nuh'],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                check=True,
            ),
        ]

        run_mock.side_effect = CalledProcessError(1, '')
        self.assertListEqual(docker_images_check(), [
            Error(
                'Docker image with tag "nuh" was not found.',
                hint='settings.ARKINDEX_TASKS_IMAGE = "nuh"',
                id='arkindex.E006',
            )
        ])

        self.assertEqual(run_mock.call_count, 1)
        self.assertEqual(run_mock.call_args_list, expected_calls)

    @patch('arkindex.project.checks.subprocess.run')
    def test_docker_images_check_missing_client(self, run_mock):
        """
        Test the Docker images check does not show errors if the Docker client is missing
        """
        from arkindex.project.checks import docker_images_check

        run_mock.side_effect = FileNotFoundError
        with self.settings(ARKINDEX_APP_IMAGE='nope', ARKINDEX_TASKS_IMAGE='nuh'):
            self.assertListEqual(docker_images_check(), [])

        self.assertEqual(run_mock.call_count, 1)
        self.assertEqual(run_mock.call_args_list, [
            call(
                ['docker', 'image', 'inspect', 'nuh'],
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                check=True,
            ),
        ])

    @override_settings()
    @patch('arkindex.project.checks.parse_recipe')
    def test_ponos_recipe_check(self, parse_mock):
        from arkindex.project.checks import ponos_recipe_check
        self.assertListEqual(ponos_recipe_check(), [])

        parse_mock.side_effect = AssertionError('oh no')
        self.assertListEqual(ponos_recipe_check(), [Error(
            'Base Ponos recipe validation failed: oh no',
            hint='settings.PONOS_RECIPE',
            id='arkindex.E007',
        )])

        settings.CONFIG_PATH = Path('/somewhere/config.yml')
        del settings.PONOS_RECIPE['env']
        parse_mock.side_effect = None
        self.assertListEqual(ponos_recipe_check(), [
            Warning(
                'The ARKINDEX_API_URL environment variable should be defined '
                'to allow API client autoconfiguration in Ponos tasks',
                hint='`ponos.default_env` in /somewhere/config.yml',
                id='arkindex.W006',
            ),
            Warning(
                'The ARKINDEX_API_TOKEN environment variable should be defined '
                'to allow API client autoconfiguration in Ponos tasks',
                hint='`ponos.default_env` in /somewhere/config.yml',
                id='arkindex.W006',
            ),
            Warning(
                'The ARKINDEX_API_CSRF_COOKIE environment variable should be defined '
                'to allow API client autoconfiguration in Ponos tasks',
                hint='`ponos.default_env` in /somewhere/config.yml',
                id='arkindex.W006',
            ),
        ])

    @override_settings()
    def test_gitlab_oauth_check(self):
        from arkindex.project.checks import gitlab_oauth_check

        del settings.GITLAB_APP_ID
        del settings.GITLAB_APP_SECRET

        self.assertListEqual(gitlab_oauth_check(), [
            Warning(
                'GitLab app ID is not set; Git imports will fail.',
                hint='settings.GITLAB_APP_ID = None',
                id='arkindex.W003',
            ),
            Warning(
                'GitLab app secret is not set; Git imports will fail.',
                hint='settings.GITLAB_APP_SECRET = None',
                id='arkindex.W004',
            ),
        ])

        settings.GITLAB_APP_ID = '1234'
        settings.GITLAB_APP_SECRET = 's3kr3t'

        self.assertListEqual(gitlab_oauth_check(), [])

    @override_settings()
    def test_s3_check(self):
        from arkindex.project.checks import s3_check

        del settings.AWS_ACCESS_KEY
        del settings.AWS_SECRET_KEY
        del settings.AWS_THUMBNAIL_BUCKET
        del settings.AWS_STAGING_BUCKET
        self.assertCountEqual(s3_check(), [
            Error(
                'AWS access key ID is missing; all S3-related features will fail.',
                hint='settings.AWS_ACCESS_KEY = None',
                id='arkindex.E011',
            ),
            Error(
                'AWS secret key is missing; all S3-related features will fail.',
                hint='settings.AWS_SECRET_KEY = None',
                id='arkindex.E011',
            ),
            Error(
                'S3 thumbnails bucket name is missing; all S3-related features will fail.',
                hint='settings.AWS_THUMBNAIL_BUCKET = None',
                id='arkindex.E011',
            ),
            Error(
                'S3 staging bucket name is missing; all S3-related features will fail.',
                hint='settings.AWS_STAGING_BUCKET = None',
                id='arkindex.E011',
            ),
        ])

        settings.AWS_ACCESS_KEY = 'key'
        settings.AWS_SECRET_KEY = 's3kr3t'
        settings.AWS_THUMBNAIL_BUCKET = 'Thumbs.db'
        settings.AWS_STAGING_BUCKET = 'buckette'
        self.assertListEqual(s3_check(), [])

    @override_settings()
    def test_public_hostname_check(self):
        from arkindex.project.checks import public_hostname_check

        settings.PUBLIC_HOSTNAME = None
        self.assertListEqual(public_hostname_check(), [
            Warning(
                'The public_hostname setting is missing; absolute URLs may not be correctly generated.',
                hint='settings.PUBLIC_HOSTNAME',
                id='arkindex.W007',
            )
        ])

        settings.PUBLIC_HOSTNAME = 'https://darkindex.lol'
        self.assertListEqual(public_hostname_check(), [])
