import urllib.parse
from unittest.mock import MagicMock, patch

from django.http.request import HttpRequest
from django.urls import reverse

from arkindex.project.tests import FixtureTestCase
from arkindex.users.models import OAuthStatus
from arkindex.users.providers import GitLabOAuthProvider


class TestGitLabOAuthProvider(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.gl_patch = patch('arkindex.users.providers.Gitlab')

    def setUp(self):
        super().setUp()
        self.gl_mock = self.gl_patch.start()

    def tearDown(self):
        self.gl_patch.stop()
        super().tearDown()

    def test_enabled(self):
        with self.settings(GITLAB_APP_ID=None, GITLAB_APP_SECRET=None):
            self.assertFalse(GitLabOAuthProvider.enabled())
        with self.settings(GITLAB_APP_ID='abcd', GITLAB_APP_SECRET='1234'):
            self.assertTrue(GitLabOAuthProvider.enabled())

    def test_callback_uri(self):
        request_mock = MagicMock(spec=HttpRequest)
        GitLabOAuthProvider(request=request_mock).get_callback_uri()
        self.assertEqual(request_mock.build_absolute_uri.call_count, 1)
        args, kwargs = request_mock.build_absolute_uri.call_args
        self.assertTupleEqual(args, (reverse('api:oauth-callback', kwargs={'provider': 'gitlab'}), ))
        self.assertDictEqual(kwargs, {})

    def test_authorize_uri(self):
        request_mock = MagicMock(spec=HttpRequest)
        request_mock.build_absolute_uri.return_value = 'callback'
        with self.settings(GITLAB_APP_ID='abcd'):
            uri = GitLabOAuthProvider(request=request_mock, credentials=self.creds).get_authorize_uri()
        parsed = urllib.parse.urlparse(uri)
        self.assertEqual(parsed.scheme, 'https')
        self.assertEqual(parsed.netloc, 'somewhere')
        self.assertEqual(parsed.path, '/oauth/authorize')
        query = urllib.parse.parse_qs(parsed.query)
        self.assertDictEqual(query, {
            'client_id': ['abcd'],
            'redirect_uri': ['callback'],
            'scope': ['api'],
            'response_type': ['code'],
            'state': [str(self.creds.id)],
        })

    def test_handle_callback_no_state(self):
        request_mock = MagicMock(spec=HttpRequest)
        request_mock.user = self.user
        request_mock.GET = {}
        with self.assertRaisesMessage(ValueError, 'No state hash'):
            GitLabOAuthProvider(request=request_mock).handle_callback()

    def test_handle_callback_bad_response(self):
        request_mock = MagicMock(spec=HttpRequest)
        request_mock.user = self.user
        request_mock.GET = {'state': str(self.creds.id)}
        with self.assertRaisesRegex(ValueError, 'valid response'):
            GitLabOAuthProvider(request=request_mock).handle_callback()

    def test_handle_callback_error(self):
        request_mock = MagicMock(spec=HttpRequest)
        request_mock.user = self.user
        request_mock.GET = {'state': str(self.creds.id), 'error': 'error message'}
        with self.assertRaisesMessage(ValueError, 'error message'):
            GitLabOAuthProvider(request=request_mock).handle_callback()

    def test_handle_callback_overwrite(self):
        self.creds.status = OAuthStatus.Done
        self.creds.save()
        request_mock = MagicMock(spec=HttpRequest)
        request_mock.user = self.user
        request_mock.GET = {'state': str(self.creds.id), 'code': 'something'}
        with self.assertRaisesRegex(Exception, 'overwrite'):
            GitLabOAuthProvider(request=request_mock).handle_callback()

    @patch('arkindex.users.providers.requests')
    def test_handle_callback_success(self, requests):
        request_mock = MagicMock(spec=HttpRequest)
        request_mock.user = self.user
        request_mock.GET = {'state': str(self.creds.id), 'code': 'abc123'}
        request_mock.build_absolute_uri.return_value = 'callback'

        self.gl_mock.return_value.user.username = 'bobby'

        requests.post.return_value.json.return_value = {
            'access_token': 't0k3n',
            'refresh_token': 'r3fr3sh',
            'token_type': 'Bearer',
            'expires_in': '3600',
        }

        with self.settings(GITLAB_APP_ID='abcd', GITLAB_APP_SECRET='s3kr3t'):
            GitLabOAuthProvider(request=request_mock).handle_callback()

        self.creds.refresh_from_db()
        self.assertEqual(self.creds.token, 't0k3n')
        self.assertEqual(self.creds.refresh_token, 'r3fr3sh')
        self.assertEqual(self.creds.account_name, 'bobby')
        self.assertEqual(self.creds.status, OAuthStatus.Done)
        self.assertEqual(requests.post.call_count, 1)
        self.assertEqual(self.gl_mock.return_value.auth.call_count, 1)

        args, kwargs = requests.post.call_args
        self.assertDictEqual(kwargs, {})
        self.assertEqual(len(args), 2)
        parsed = urllib.parse.urlparse(args[0])
        self.assertEqual(parsed.scheme, 'https')
        self.assertEqual(parsed.netloc, 'somewhere')
        self.assertEqual(parsed.path, '/oauth/token')
        self.assertEqual(parsed.query, '')
        self.assertDictEqual(args[1], {
            'client_id': 'abcd',
            'client_secret': 's3kr3t',
            'code': 'abc123',
            'redirect_uri': 'callback',
            'grant_type': 'authorization_code',
        })

        args, kwargs = self.gl_mock.call_args
        self.assertEqual(args, ('https://somewhere', ))
        self.assertDictEqual(kwargs, {'oauth_token': 't0k3n'})

    def test_disconnect(self):
        request_mock = MagicMock(spec=HttpRequest)
        request_mock.build_absolute_uri.return_value = 'hook'

        hook_mock = MagicMock()
        hook_mock.url = 'hook'
        self.gl_mock.return_value.projects.get.return_value.id = 'repo_id'
        self.gl_mock.return_value.projects.get.return_value.hooks.list.return_value = [hook_mock]

        GitLabOAuthProvider(request=request_mock, credentials=self.creds).disconnect()

        # Number of repositories associated to those credentials
        repos_count = self.creds.repos.count()
        self.assertEqual(self.gl_mock().projects.get.call_count, repos_count)
        self.assertEqual(self.gl_mock().projects.get.return_value.hooks.list.call_count, repos_count)
        self.assertEqual(hook_mock.delete.call_count, repos_count)
